//////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2002-2014 Unreal Streaming Technologies
//
// URTMPClientViewer sample program that shows how to initialize and use URTMPClientSrc DirectShow Filter
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <conio.h> 
#include <initguid.h>
#include "URTMPClientSrc.h"

#define STREAM_NONE  0
#define STREAM_VIDEO 1
#define STREAM_AUDIO 2

#define STREAM_TYPE BYTE


volatile bool bBreakWait = true;


BOOL WINAPI HandlerRoutine(DWORD dwCtrlType)
{
    bBreakWait = false;
    return TRUE;
}

HRESULT GetFilter(REFCLSID clsidDeviceClass, LPCWSTR szName, IBaseFilter** pFilter)
{
    if (!pFilter)
    {
        return E_POINTER;
    }

    *pFilter = NULL;

    CComPtr<ICreateDevEnum> spSysDevEnum;
    HRESULT hr = CoCreateInstance(CLSID_SystemDeviceEnum, 
                                  NULL, 
                                  CLSCTX_INPROC_SERVER, 
                                  IID_ICreateDevEnum, 
                                  reinterpret_cast<void**>(&spSysDevEnum.p));
    if (FAILED(hr))
    {
        return hr;
    }

    CComPtr<IEnumMoniker> spEnum;    
    hr = spSysDevEnum->CreateClassEnumerator(clsidDeviceClass, &spEnum.p, 0);
    if (hr != S_OK)
    {        
        return SUCCEEDED(hr) ? HRESULT_FROM_WIN32(ERROR_NOT_FOUND) : hr;
    }

    USES_CONVERSION;

    HRESULT hrFinal = HRESULT_FROM_WIN32(ERROR_NOT_FOUND);

    CComPtr<IMoniker> spMoniker;    
    while (S_OK == spEnum->Next(1, &spMoniker.p, NULL))
    {
        CComPtr<IPropertyBag> spPropBag;
        hr = spMoniker->BindToStorage(0, 0, IID_IPropertyBag, reinterpret_cast<void**>(&spPropBag.p));
        if (FAILED(hr))
        {
            return hr;
        }

        CComVariant varName;        
        hr = spPropBag->Read(L"FriendlyName", &varName, 0);
        if (SUCCEEDED(hr))
        {
            if (lstrcmp(szName, OLE2CW(varName.bstrVal)) == 0)
            {
                hrFinal = spMoniker->BindToObject(NULL, NULL, IID_IBaseFilter, reinterpret_cast<void**>(pFilter));                
                break;
            }
        }

        spMoniker = NULL;
    }

    return hrFinal;
}

HRESULT FindSupportedStream(IPin* pPin, STREAM_TYPE& StreamType)
{    
    StreamType = STREAM_NONE;

    bool bRet = false;
    CComPtr<IEnumMediaTypes> spEnumMediaTypes;
    HRESULT hr = pPin->EnumMediaTypes(&spEnumMediaTypes);
    if (FAILED(hr))
    {
        return hr;
    }

    AM_MEDIA_TYPE* pMediaType = NULL;
    hr = spEnumMediaTypes->Next(1, &pMediaType, NULL);
    if (hr == S_OK)
    {
        if (pMediaType->pbFormat)
        {
            if (pMediaType->majortype == MEDIATYPE_Video)
            {
                StreamType = STREAM_VIDEO;
            }
            else if (pMediaType->majortype == MEDIATYPE_Audio)
            {
                StreamType = STREAM_AUDIO;
            }
        }

        DeleteMediaType(pMediaType);
        pMediaType = NULL;
    }

    return hr;
}

HRESULT BuildCaptureGraph(IBaseFilter* pRTMPSource, STREAM_TYPE StreamTypes, IGraphBuilder** ppGraph)
{    
    if (StreamTypes == STREAM_NONE)
    {
        //
        // RTMP source must support at least one video or audio
        //
        return E_UNEXPECTED;
    }

    CComPtr<IGraphBuilder> spGraph;    
    HRESULT hr = spGraph.CoCreateInstance(CLSID_FilterGraph);
    if (FAILED(hr))
    {
        return hr;
    }
    
    CComPtr<ICaptureGraphBuilder2> spGraphBuilder;
    hr = spGraphBuilder.CoCreateInstance(CLSID_CaptureGraphBuilder2);
    if (FAILED(hr))
    {
        return hr;
    }
    
    spGraphBuilder->SetFiltergraph(spGraph.p);    

    spGraph->AddFilter(pRTMPSource, L"RTMP Source");

    if (StreamTypes & STREAM_VIDEO)
    {
        CComPtr<IBaseFilter> spVideoRenderer;
        hr = GetFilter(CLSID_LegacyAmFilterCategory, CComBSTR(L"Video Renderer"), &spVideoRenderer.p);
        if (FAILED(hr))
        {
            return hr;
        }

        spGraph->AddFilter(spVideoRenderer, L"Video Renderer");

        hr = spGraphBuilder->RenderStream(NULL, &MEDIATYPE_Video, pRTMPSource, NULL, spVideoRenderer.p);
        if (FAILED(hr))
        {
            return hr;
        }
    }

    if (StreamTypes & STREAM_AUDIO)
    {
        CComPtr<IBaseFilter> spAudioRenderer;
        hr = GetFilter(CLSID_AudioRendererCategory, CComBSTR(L"Default DirectSound Device"), &spAudioRenderer.p);
        if (FAILED(hr))
        {
            return hr;
        }

        spGraph->AddFilter(spAudioRenderer, L"Audio Renderer");

        hr = spGraphBuilder->RenderStream(NULL, &MEDIATYPE_Audio, pRTMPSource, NULL, spAudioRenderer.p);
        if (FAILED(hr))
        {
            return hr;
        }
    }

    *ppGraph = spGraph.Detach();

    return hr;
}

HRESULT PlayRTMP(LPCWSTR szInitString)
{
    //
    // Create instance of Unreal RTMP Client Source filter
    //
    CComPtr<IBaseFilter> spRTMPSource;
    HRESULT hr = GetFilter(CLSID_VideoInputDeviceCategory, L"Unreal RTMP Client Source", &spRTMPSource.p);    
    if (FAILED(hr))
    {
        return hr;
    }

    //
    // Get IFileSourceFilter interface to initialize the filter 
    //
    CComQIPtr<IFileSourceFilter> spFileSource = spRTMPSource;
    if (!spFileSource.p)
    {
        return E_UNEXPECTED;
    }

    //
    // Initialize Unreal RTMP Client Source filter with initialization string
    //
    hr = spFileSource->Load(szInitString, NULL);
    if (FAILED(hr))
    {
        return hr;
    }

    //
    // Now, after initialization succeeded output pins must expose media types.
    // Enumerate the pins and detect media types RTMP source supports.
    //
    STREAM_TYPE StreamTypes = STREAM_NONE;

    CComPtr<IEnumPins> spEnum;
    CComPtr<IPin>      spPin;
    hr = spRTMPSource->EnumPins(&spEnum.p);
    if (FAILED(hr))
    {
        return hr;
    }    

    while (spEnum->Next(1, &spPin.p, 0) == S_OK)
    {
        PIN_DIRECTION PinDir;
        hr = spPin->QueryDirection(&PinDir);
        if (FAILED(hr))
        {
            return hr;
        }

        if (PinDir == PINDIR_OUTPUT)
        {            
            STREAM_TYPE StreamTypeOnPin = STREAM_NONE;
            hr = FindSupportedStream(spPin.p, StreamTypeOnPin);
            if (FAILED(hr))
            {
                return hr;
            }

            StreamTypes |= StreamTypeOnPin;
        }

        spPin = NULL;
    }

    //
    // Build capture graph
    //
    CComPtr<IGraphBuilder> spGraph;
    hr = BuildCaptureGraph(spRTMPSource.p, StreamTypes, &spGraph.p);
    if (FAILED(hr))
    {
        return hr;
    }

    CComQIPtr<IMediaEvent> spMediaEvent = spGraph;
    if (!spMediaEvent.p)
    {
        return E_UNEXPECTED;
    }

    CComQIPtr<IMediaControl> spMediaControl = spGraph;
    if (!spMediaEvent.p)
    {
        return E_UNEXPECTED;
    }

    //
    // Start the graph
    //
    hr = spMediaControl->Run();
    if (FAILED(hr))
    {
        return hr;
    }

    hr = S_OK;

    long     evCode = 0;
    LONG_PTR param1, param2;

    while (bBreakWait)
    {
        if (spMediaEvent->GetEvent(&evCode, &param1, &param2, 30) == E_ABORT)
        {
            MSG msg;
            if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
            {                
                DispatchMessage(&msg);	    
            }

            continue;
        }

        spMediaEvent->FreeEventParams(evCode, param1, param2);

        switch (evCode)
        {            
            case EC_COMPLETE:
            case EC_ERRORABORT:
            case EC_RTMP_SOURCE_ERROR:            // Custom URTMPClientSrc event id. Get error code from param1
                hr = static_cast<HRESULT>(param1);
                bBreakWait = false;
                break;

            case EC_USERABORT:
                bBreakWait = false;
                break;
        }
    } 

    spMediaControl->Stop();

    return hr;
}

int _tmain(int argc, _TCHAR* argv[])
{
    if (argc != 2)
    {
        _putts(_T("\nUsage: URTMPClientViewer \"InitString\"\n"));
        _putts(_T("Initialization String format:"));
        _putts(_T("Set of <Parameter>=<value> pairs separated by '|' charachter"));
        _putts(_T("Parameters:"));
        _putts(_T("URL:          RTMP source URL. Must start with 'RTMP://'"));
        _putts(_T("BufferLength: Pre-buffering in seconds if present, range: 0-10 (default is 0)"));
        _putts(_T("\nExample of initialization string:"));
        _putts(_T("\"URL = rtmp://cwedge.cdcache.net/cwedge/rilive | BufferLength = 2\"\n"));

        _getch();
        return 0;
    }

    SetConsoleCtrlHandler(HandlerRoutine, TRUE);

    CoInitialize(NULL);
    
    USES_CONVERSION;
    HRESULT hr = PlayRTMP(T2CW(argv[1]));
    if (FAILED(hr))
    {
        if (hr == VFW_E_CANNOT_CONNECT)
        {
            _putts(_T("Your system seems to be missing decoder or intermediate DirectShow filter"));
        }
        else if (hr == HRESULT_FROM_WIN32(ERROR_EVALUATION_EXPIRATION))
        {
            _putts(_T("Demo version play time limit reached. Restart the application to continue evaluation."));
        }
        else
        {
            wprintf(L"\nPlay error 0x%08X occurred\n", hr);
        }
    }

    CoUninitialize();

    SetConsoleCtrlHandler(HandlerRoutine, FALSE);

	return 0;
}